// Kristen Kleckner - Capstone 2015
// Adapted from Dave Ferris's 2014 Capstone
// Last Modified May 10, 2015

#include <fstream>
#include <iostream>
#include <string>
#include <cmath>
#include <sys/stat.h>
#include "Word.h"
#include "Hash.h"
#include "Heap.h"
#include "TweetInfo.h"
using namespace std;

//---------------------- Prototypes ----------------------
int DefineGlobalArrays();
int AddWordsFromSentence(int index, Hash & word_table);
int WriteHashResultsToFile(Hash * words, int sentences_count, string output_filename);
int WriteHeapResultsToFile(Heap<TweetInfo> * heap, string output_filename);
int LowercaseString(string & text);
int LowercaseHashtag(string & hashtag);
bool IsIgnoredWord(string word);
int GetFileData();
int FindNearestPrime(int beginning_value);
bool IsPrime(int val);
//bool TitleDetected(string sentence);
void ParseTweets(string line, int index, Hash & word_table, Hash & hashtag_table, Hash & user_table, Hash & mentions_table, Heap<TweetInfo> & favorites_table, Heap<TweetInfo> & retweets_table);
string cstrtok(string &data, string delims);
void QuickSort(int A[], string B[], int first, int last);
int Split(int Count[], string Text[], int first, int last);

//------------------- Global Variables -------------------
int CHAR_ENDS_SENTENCE [256];							// Signifies whether each char (ascii 0-255) is sentence-ending
int SPECIAL_CHARS [256];								// Signifies whether each char (ascii 0-255) is special (i.e. not alphabetical)
int IGNORED_WORDS_COUNT;								// Tracks the size of the IGNORED_WORDS array.
string * IGNORED_WORDS;									// Points to the array of words that we want to ignore.
// The file containing these words was taken from Michael Klosiewski's 2013 project and modified
const int TOP_WORDS = 15;								// Specifies how many words and their data the user wants to be written to the resulting text file.
string EARLIEST_DATE;									// Strings to hold date so it can be written to an output file
string LATEST_DATE;

int main()
{
	if (DefineGlobalArrays() == -1)
	{
		cout << "ERROR: DefineGlobalArrays failed. Exiting.." << endl;
		return 0;
	}

	// Since we want the Hash size to be nearly equal to the expected amount of items, and prime, according to Dr. McVey, we will
	// find the approx. amount of words, and then find the nearest prime number >= the word estimate 
	int total_tweets = GetFileData();

	// Find the nearest prime #, >= to the calculated size
	int calculated_hash_size = FindNearestPrime(total_tweets * 5);	//was set to 20, but had so much empty space that I cut it down significantly (6 letters is average, but we exclude common words so we can definitely cut it down)
	int calculated_user_hash_size = FindNearestPrime(total_tweets);

	// Dumping data
	//cout << "Word Estimate: " << file_word_estimate << endl;
	//cout << "Size for hash: " << calculated_hash_size << endl;
	cout << "Tweets to analzye: " << total_tweets << endl;
	cout << "Size for hash: " << calculated_hash_size << endl;
	cout << "Size for user hash: " << calculated_user_hash_size << endl;	//user hash size is different since we know there are significantly less users than words

	// Create the hash tables and heaps
	Hash word_table = Hash(calculated_hash_size);
	Hash hashtag_table = Hash(calculated_hash_size);
	Hash user_table = Hash(calculated_user_hash_size);
	Hash mentions_table = Hash(calculated_hash_size);

	//heaps can get the same value as hash tables since they all need to have roughly the same space
	Heap<TweetInfo> favorites_heap = Heap<TweetInfo>(total_tweets);
	Heap<TweetInfo> retweets_heap = Heap<TweetInfo>(total_tweets);

	//open the input data file
	ifstream tweet_file("Data/input.txt");

	if(!tweet_file.is_open())
	{
		cout << "ERROR: File could not open. Exiting..." << endl;
		return 0;
	}

	//initialize variables
	string line, tweet;
	int total_tweet_count = 0;
	while(!tweet_file.eof())
	{
		getline(tweet_file, line);
		total_tweet_count++; //increment tweet count first so the line numbers stay accurate.  if count is incremented after, the line numbers in the hashtable/heap are off by 1
		ParseTweets(line, total_tweet_count, word_table, hashtag_table, user_table, mentions_table, favorites_heap, retweets_heap);
		
		/**********************************
		For testing purposes:
		Sometimes, a hashtable will automatically get millions of entries without any logical reason
		Uncomment this code to find the line number where this occurs
		Remove that line from input.txt and tweets.txt and rerun the program
		This should fix the issue
		***********************************/
		//if(user_table.GetTotalWords() > 0)
		//	cout << total_tweet_count << " - " << user_table.GetTotalWords() << endl;
		//user_table.ModifiedDump();
	}
	
	//close the tweet file
	tweet_file.close();

	//output to files
	if (WriteHeapResultsToFile(&retweets_heap, "Data/retweets.txt") == -1)
	{
		cout << "ERROR: WriteHeapResultsToFile failed. Exiting.." << endl;
		return 0;
	}

	if (WriteHeapResultsToFile(&favorites_heap, "Data/favorites.txt") == -1)
	{
		cout << "ERROR: WriteHeapResultsToFile failed. Exiting.." << endl;
		return 0;
	}

	// Check to see if WriteHashResultsToFile for top words errored
	if (WriteHashResultsToFile(&word_table, total_tweet_count, "Data/words.txt") == -1)
	{
		cout <<	"ERROR: WriteHashResultsToFile failed. Exiting.." << endl;
		return 0;
	}

	// Check to see if WriteHashResultsToFile for top users errored
	if (WriteHashResultsToFile(&user_table, total_tweet_count, "Data/users.txt") == -1)
	{
		cout <<	"ERROR: WriteHashResultsToFile failed. Exiting.." << endl;
		return 0;
	}

	// Check to see if WriteHashResultsToFile for top hashtags errored
	if (WriteHashResultsToFile(&hashtag_table, total_tweet_count, "Data/hashtags.txt") == -1)
	{
		cout <<	"ERROR: WriteHashResultsToFile failed. Exiting.." << endl;
		return 0;
	}

	// Check to see if WriteHashResultsToFile for top mentions errored
	if (WriteHashResultsToFile(&mentions_table, total_tweet_count, "Data/mentions.txt") == -1)
	{
		cout <<	"ERROR: WriteHashResultsToFile failed. Exiting.." << endl;
		return 0;
	}

	//open the information file
	ofstream info_file("Data/information.txt");
	if(!info_file.is_open())
	{
		cout << "ERROR: File could not open. Exiting..." << endl;
		return 0;
	}

	//write the data to the information file
	info_file << IGNORED_WORDS[0] << endl;
	info_file << total_tweets << endl;
	info_file << EARLIEST_DATE << endl;
	info_file << LATEST_DATE << endl;
	info_file << word_table.GetTotalWords() << endl;
	info_file << hashtag_table.GetTotalWords() << endl;
	info_file << user_table.GetTotalWords() << endl;
	info_file << mentions_table.GetTotalWords() << endl;

	//close the file
	info_file.close();

	//user_table.ModifiedDump();

	return 0;
}

/************************************************************
Name: DefineGlobalArrays
Parameters: None
Returns: 0 if everything worked, -1 if something went wrong
Purpose: Set up the IGNORED_WORDS array to contain all the words in the 
	ignored_words text file as well as the word that was searched on,
	which is stored as the first line in information.txt

Initially created by Dave Ferris
************************************************************/
int DefineGlobalArrays()
{
	// Declare and initialize variables
	ifstream ignored_words_file;
	string filename = "ignored_words.txt";
	string line = "";
	int line_count = 0;
	string * temp;

	// Allocate an IGNORED_WORDS array of size 0, to avoid creating more cases within the loop
	IGNORED_WORDS = new string [0];

	//Open information.txt to get the word that was searched on
	ifstream info_file("Data/information.txt");
	if(!info_file.is_open())
	{
		cout << "ERROR: File could not open. Exiting..." << endl;
		return -1;
	}

	//read the first line in the file; this is the value that was searched on
	getline(info_file, line);

	//increment line_count which contains the number of words that will be ignored
	line_count++;

	//adjust line so that whatever was searched on is now lowercase
	if(line[0] == '#')
		LowercaseHashtag(line);
	else
		LowercaseString(line);

	//Allocate memory for the temp array
	temp = new string [line_count];

	// Transfer data from IGNORED_WORDS to temp
	for(int i = 0; i < (line_count - 1); i++)
		temp[i] = IGNORED_WORDS[i];

	// Add the newest line to the temp array
	temp[line_count - 1] = line;

	// Delete the old IGNORED_WORDS ARRAY
	delete [] IGNORED_WORDS;

	// Point IGNORED_WORDS to temp
	IGNORED_WORDS = temp;

	// Point temp to null
	temp = 0;

	//close the information file
	info_file.close();

	//open the ignored words file
	ignored_words_file.open(filename);

	if(!ignored_words_file.is_open())
	{
		std::cout << "ERROR: ignored words file could not be opened. Exiting.." << endl;
		return -1;
	}

	while(1)
	{
		// Get the next word from the file
		getline(ignored_words_file, line);
		line_count++;

		// If the end of the file was reached, break the loop
		if(ignored_words_file.eof())
			break;

		// Allocate memory for the temp array
		temp = new string [line_count];

		// Transfer data from IGNORED_WORDS to temp
		for(int i = 0; i < (line_count - 1); i++)
			temp[i] = IGNORED_WORDS[i];

		// Add the newest line to the temp array
		temp[line_count - 1] = line;

		// Delete the old IGNORED_WORDS ARRAY
		delete [] IGNORED_WORDS;

		// Point IGNORED_WORDS to temp
		IGNORED_WORDS = temp;

		// Point temp to null
		temp = 0;
	}

	

	IGNORED_WORDS_COUNT = line_count - 1;

	// Test dump the IGNORED_WORDS array
	/*for(int j = 0; j < (line_count - 1); j++)
	{
		cout << j << ": " << IGNORED_WORDS[j] << endl;
	}*/

	return 0;
}

/************************************************************
Name: WriteHashResultsToFile
Parameters: Hash table, the number of tweets, and the name of the file
Returns: 0 if everything worked properly, -1 if something went wrong
Purpose: Record the information from the given hash table to a file. 
	This file can then be used in the C# application to visualize data.

Intially created by Dave Ferris
************************************************************/
int WriteHashResultsToFile(Hash * words, int sentences_count, string output_filename)
{
	// Gets the total amount of words in the words hash
	const int TOTAL_WORD_COUNT = words->GetTotalWords();

	// Declare new variables
	string * word_text = new string[TOTAL_WORD_COUNT];
	int * word_occurrances = new int[TOTAL_WORD_COUNT];
	int * sentence_indices = 0;
	int sentence_occurrance_count = 0;
	string sentence = "";
	ofstream file_output;

	//Get information from the hashtable (the word, number of occurrances, and instances of the word)
	words->GetWordData(word_text, word_occurrances);

	// Sends the two arrays to be sorted by occurrance count
	// Both must be sent in order to remain parallel
	QuickSort(word_occurrances, word_text, 0, TOTAL_WORD_COUNT-1);

	// open the output file
	file_output.open(output_filename);

	// If the file was not opened, display error message and return
	if(!file_output.is_open())
	{
		cout << "ERROR: Output file could not be opened -- Program terminating" << endl;
		return -1;
	}

	// If the file was opened, display confirmation
	else
	{
		cout << "File " << output_filename << " was successfully opened" << endl;

		//Find out how many top words
		//If there were less than 15 words, set the number of words to be written to that number
		int numWords;
		if(TOTAL_WORD_COUNT < TOP_WORDS)
			numWords = TOTAL_WORD_COUNT;
		else
			numWords = TOP_WORDS;

		// Writes the top numWords most used words to the file - one per line
		for(int j = 0; j < numWords; j++)
		{
			//Write the word
			file_output << word_text[j] << "\t";

			//Get the list of all of the instances of the word
			sentence_indices = words->FindWordByText(word_text[j])->GetSentenceIndices();
			sentence_occurrance_count = words->FindWordByText(word_text[j])->GetAppearancesCount();

			// Prints all but the last sentence index
			for(int z = 0; z < (sentence_occurrance_count - 1); z++)
				file_output << sentence_indices[z] << ",";

			// Prints the final sentence index, with no trailing comma
			file_output << sentence_indices[sentence_occurrance_count - 1] << "\n";
		}

		// Close the file
		file_output.close();

		//Display a message that says that it was successful
		cout << "File " << output_filename << " was successfully closed" << endl;
		return 0;
	}
}

/************************************************************
Name: WriteHeapResultsToFile
Parameters: Heap and the name of the output file
Returns: 0 if everything worked properly, -1 if something went wrong
Purpose: Record the information from the given heap to a file. 
	This file can then be used in the C# application to visualize data.
************************************************************/
int WriteHeapResultsToFile(Heap<TweetInfo> * heap, string output_filename)
{
	//Open the file for writing
	ofstream heap_file;
	heap_file.open(output_filename);

	if(!heap_file.is_open())
	{
		cout << "ERROR: File could not open. Exiting..." << endl;
		return -1;
	}

	//Determine how many tweets were analyzed
	//If it was less that TOP_WORDS, only run that many
	int number_of_tweets;
	if(heap->GetTotalElements() < TOP_WORDS)
		number_of_tweets = heap->GetTotalElements();
	else
		number_of_tweets = TOP_WORDS;

	//Extract the max of the heap number_of_tweets times 
	//Write the max to the file
	for(int i = 0; i < number_of_tweets; i++)
	{
		heap_file << heap->ExtractMax() << endl;
	}

	//Close the file
	heap_file.close();

	return 0;
}

/************************************************************
Name: LowercaseString
Paramters: a string to convert
Returns: -1 if something went wrong (if there is a non-ascii value in the string)
		  0 if everything worked
		  the converted string
Purpose: Convert all the characters of a string to lowercase to help with string
	comparisons.  It will fail if non-ascii values are found in the string

Initially written by Dave Ferris
************************************************************/
int LowercaseString(string & text)
{
	for(int i = 0; i < text.length(); i++)
	{	
		if(isascii(text[i]))
		{
			text[i] = tolower(text[i]);
		}
		else
			return -1;
	}

	return 0;
}

/************************************************************
Name: LowercaseHashtag
Paramters: a string to convert
Returns: -1 if something went wrong (if there is a non-ascii value in the string)
		  0 if everything worked
		  the converted string
Purpose: Convert all the characters of a string to lowercase to help with string
	comparisons.  It will fail if non-ascii values are found in the string

Adapted from LowercaseString
************************************************************/
int LowercaseHashtag(string & hashtag)
{
	//start at i = 1 so we leave the actual # intact
	for(int i = 1; i < hashtag.length(); i++)
	{
		if(isascii(hashtag[i]))
		{
			hashtag[i] = tolower(hashtag[i]);
		}
		else
			return -1;
	}

	return 0;
}

/************************************************************
Name: IsIgnoredWord
Parameters: a string
Returns: true or false, depending on if the word is in IGNORED_WORDS array
Purpose: Determine if the given word is in the IGNORED_WORDS array.
	If it is, we need to make sure it does not get entered into the Hash Table

Initially written by Dave Ferris
************************************************************/
bool IsIgnoredWord(string word)
{
	for (int i = 0; i < IGNORED_WORDS_COUNT; i++)
	{
		if (word == IGNORED_WORDS[i])
			return true;
	}

	return false;
}

/************************************************************
Name: GetFileData
Parameters: None
Returns: the number of tweets that were gathered
Purpose: Read the information.txt file to get data.
	We can set the EARLIEST_DATE and LATEST_DATE which need
	to be written later.
	We also get the number of tweets that were analyzed which
	help us make better estimations as to how much space we need
	to allocate for the hash tables and heaps
************************************************************/
int GetFileData()
{
	ifstream info_file("Data/information.txt");
	if(!info_file.is_open())
	{
		cout << "ERROR: File could not open. Exiting..." << endl;
		return 0;
	}

	string line;

	getline(info_file, line);	//this is value that was searched on; it is not needed here
	getline(info_file, line);	//this line has the total number of tweets
	int tweet_count = stoi(line);
	getline(info_file, line);
	EARLIEST_DATE = line;
	getline(info_file, line);
	LATEST_DATE = line;

	info_file.close();



	return tweet_count;
}

/************************************************************
Name: FindNearestPrime
Parameters: some integer that is the estimation for the size of the hash table
Returns: the prime number closest to the given integer
Purpose: Find the first prime number greater than or equal to 
	the given integer so we can get a good hash table size.

Initially written by Dave Ferris
************************************************************/
int FindNearestPrime(int beginning_value)
{
	int val = beginning_value;

	//if the value is even, add one to make it odd
	if(val%2 == 0)
		val++;

	// Infinite while - should be broken by return statement
	while(1)
	{
		if(IsPrime(val))
			return val;
		else
			val += 2;
	}
}

/************************************************************
Name: IsPrime
Parameters: an integer
Returns: true or false, depending on if the given int is prime
Purpose: Determine where the given int is a prime number

Initially written by Dave Ferris
************************************************************/
bool IsPrime(int val)
{
	// It is only necessary to go to (val/2) - 1, and all possibilities
	//  will have been covered. If 2 is not a factor of val, then val/2
	//  is not a factor either
	for(int i = 2; i < val/2; i++)
	{
		if((val % i) == 0)
			return false;
	}

	return true;
}

/************************************************************
Name: ParseTweets
Parameters: a tweet, the line number corresponding to that tweet, all four hash tables and both heaps
Returns: Modified hash tables and heaps
Purpose: Analyze the tweet that was passed in by parsing the data on tabs.
	Then we send the data to the appropriate hash tables and heaps
Note: Tweets are formatted as follows:
	Date \t User \t Retweets \t Favorites \t Tweet Text
************************************************************/
void ParseTweets(string line, int i, Hash & word_table, Hash & hashtag_table, Hash & user_table, Hash & mentions_table, Heap<TweetInfo> & favorites_heap, Heap<TweetInfo> & retweets_heap)
{
	//declare a new string so we don't have to change the initial string that we were sent
	string tweet;

	//if the string is empty, there is nothing to analyze
	if(line == "")
		return;

	//Use cstrtok to split on the tab
	//The first piece of data is the date
	//We don't currently do anything with the date so we can ignore it
	tweet = cstrtok(line,"\t");
	if(tweet == "")
	{
		cout << "ERROR: Problem with tweet file - date - line " << i << endl;
		return;
	}
	
	//Split the line again to get the user
	tweet = cstrtok(line,"\t");
	if(tweet == "")
	{
		cout << "ERROR: Problem with tweet file - user - line " << i << endl;
		return;
	}
	
	//Insert into the user hash table
	//cout << "Inserting into user table: " << tweet << endl;
	user_table.Add(Word(tweet,i));

	//Split on the tab to get the number of retweets
	tweet = cstrtok(line,"\t");
	if(tweet == "")
	{
		cout << "ERROR: Problem with tweet file - retweets - line " << i << endl;
		return;
	}
	
	//Convert the value to a int
	int retweets = stoi(tweet);
	//cout << "Inserting into retweets heap: " << retweets << endl;

	//Don't insert retweets into heaps right now.
	//If the tweet is a retweet itself, the user will get credit for the initial tweet

	//Split on the tab to get the number of favorites
	tweet = cstrtok(line,"\t");
	if(tweet == "")
	{
		cout << "ERROR: Problem with tweet file - favorites - line " << i << endl;
		return;
	}
	
	//Convert the value to an int	
	int favorites = stoi(tweet);
	//cout << "Inserting into favorites heap: " << favorites << endl;

	//Don't insert retweets into heaps right now.
	//If the tweet is a retweet itself, the user will get credit for the initial tweet	

	//Split on tab to get the tweet text
	tweet = cstrtok(line,"\t");
	if(tweet == "")
	{
		cout << "ERROR: Problem with tweet file - text - line " << i << endl;
		return;
	}

	//cout << "*******Analyzing Tweet Text*******" << endl;

	//Create another string variable so we don't mess up the tweet text
	string tweetText;

	do
	{
		//cout << "text to parse: " << tweet << endl;
		if(tweet == "")
		{
			//cout << "nothing to hash...breaking out." << endl;
			break;
		}

		//Split the text on the space to get individual words
		tweetText = cstrtok(tweet, " ");

		if(tweetText == "")
		{
			//cout << "nothing to hash...breaking out." << endl;
			break;
		}
		
		//If we find RT or (rt), the tweet is a retweet
		//We don't want to analyze any more of this tweet
		//Break out of this loop so we can move on 
		if(!tweetText.find("RT") || !tweetText.find("(rt)"))
		{
			//cout << "found retweet: " << tweetText << endl;
			break;
		}

		//If the first character of the string is a #
		//We found a hash tag
		if(tweetText[0] == '#')
		{
			//cout << "Found hashtag: " << tweetText << endl;

			//Convert so the hashtag is all lowercase
			if(LowercaseHashtag(tweetText) != -1)
			{
				//If the hash tag is not ignored, hash the hash tag
				if(!IsIgnoredWord(tweetText))
					hashtag_table.Add(Word(tweetText,i));
			}
		}
		//If the first character is an @
		//We found a mention
		else if(tweetText[0] == '@')
		{
			//cout << "Found mention: " << tweetText << endl;
		
			//If it is not ignored, hash the username
			if(!IsIgnoredWord(tweetText))
				mentions_table.Add(Word(tweetText, i));
		}
		//If the string contains http, it is a URL and we can ignore it
		else if(!tweetText.find("http"))
		{
			//cout << "Found link: " << tweetText << endl;
		}
		//Otherwise, we have found a word
		else
		{
			//cout << "Hashing word: " << tweetText << endl;
			
			//Convert it to lowercase
			if(LowercaseString(tweetText) != -1)
			{
				//If the word is not ignored, hash it
				if(!IsIgnoredWord(tweetText))
					word_table.Add(Word(tweetText, i));
			}
		}

	}while(true);

	//Insert into heaps after the rest of the tweet is processed to make sure the tweet is not a retweet
	retweets_heap.Insert(TweetInfo(i, retweets), i);
	favorites_heap.Insert(TweetInfo(i, favorites), i);
	
	//cout << "****************************" << endl;
}

//from Dr. Blahnik
//Used for parsing the tweets
string cstrtok(string &data, string delims) 
{
	//***************************************************************************
	//   Returns the next token of "data".  Tokens are delimited by chars
	//   in "delims".  Note that "data" is altered on return.  (The token
	//   returning is removed from "data".)
	//   Returns "" if no more tokens in "data".
	//
	//     Typical usage:     do {
	//                           token = cstrtok(data, " ./");
	//                           if (token == "") break;  // no more
	//                           cout << token;
	//                        } while (true);
	//***************************************************************************
	int i,j;
	string tok;
	bool Adelim;

	tok = "";

	//   skip leading delims
	for(i=0; i<data.length(); i++) 
	{
		Adelim = false;
		for (j=0; j<delims.length(); j++) 
		{
			if (data[i] == delims[j]) 
			{
				Adelim = true;
				break;
			}
		}
		if (! Adelim) break;  //   beginning of the token found
	}

	//   look for 1st terminating delimiter and build token as you go
	for(; i<data.length(); i++) 
	{
		for (j=0; j<delims.length(); j++) 
		{
			if (data[i] == delims[j])      //   delim found.  End of token                                                         
			{
				data.erase(0,i);
				return (tok);
			}
		}
		tok = tok + data[i];
	}						    //   end-of-string found.  End of token.

	data.erase(0,i);
	return (tok);
} 

//Functions for QuickSort and Split, provided by Dave Ferris
/************************************************************
---------------------QuickSort and Split---------------------
	Both of these functions were taken from Dr. McVey's
	CS220 Lab #8, from spring 2012. 
	
	I have made slight modifications to cause the functions
	to sort two parallel arrays - one for text and one 
	for count. These modifications include:
	    - Changing both functions to accept another array
		- Changing variable names to discern between
		  occurrances count and text arrays
		- Adding a second pivot to track the text at the
		  index of the occurrance count pivot
		- Swapping contents of both arrays to maintain
		  parallel nature
		- Changing sort from ascending to descending

	I have chosen to use QuickSort because of its
	O(nlgn) runtime, much more time-friendly than the
	O(n^2) runtime of bubble or insertion sort. This could
	be important as I begin to work with larger text files.
************************************************************/
void QuickSort(int Count[], string Text[], int first, int last)
{
	int position;
	if (first < last)
	{
		position = Split(Count, Text, first, last);
		QuickSort(Count, Text, first, position-1);
		QuickSort(Count, Text, position+1, last);
	}
}

int Split(int Count[], string Text[], int first, int last)
{
	int count_pivot = Count[first];
	string text_pivot = Text[first];
	int left = first, right = last;

	while (left < right)
	{
		while (count_pivot > Count[right])
			right--;
		while (left < right && Count[left] >= count_pivot)
			left++;
		if (left < right)
		{
			// Whenever two cells in the Count array swap, the corresponding cells
			//  in the Text array must also, in order to keep them parallel
			swap(Count[left], Count[right]);
			swap(Text[left], Text[right]);
		}
	}

	int position = right;
	Count[first] = Count[position];
	Text[first] = Text[position];
	Count[position] = count_pivot;
	Text[position] = text_pivot;
	return position;
}